# coding: utf-8
import os
import json
from odps.models import TableSchema, Column
from odps import options
try:
  from . import pyfg
except:
  import pyfg

def get_default_value(cur_type: str):
    if cur_type.startswith('MAP') or cur_type.startswith('map'):
        return {}
    elif cur_type.startswith('ARRAY') or cur_type.startswith('array'):
        return []
    else:
        return None


class TableCreator(object):
    def __init__(self, input_table, output_table, fg_json_file, force=False):
        self.input_table = input_table
        self.output_table = output_table
        self.fg_json_file = fg_json_file
        self.force = force
        self._reserve_names = []
        self._input_names = []
        self._default_values = []
        self._output_names = []
        self._output_types = []
        self._custom_resources = []
        self._input_partitions = []

    def __call__(self, odps):
        if self.force:
            print("delete output table:", self.output_table)
            odps.delete_table(self.output_table, if_exists=True)

        if not odps.exist_resource(self.fg_json_file):
            print("[ERROR] fg json resource does not exist")
            return False
        with odps.open_resource(self.fg_json_file, mode='r') as fp:
            features = json.load(fp)

        if "reserves" in features:
            reserve_columns = set(features["reserves"])
        else:
            reserve_columns = set()

        assert "features" in features, "features not found in json file"
        cur_features = features["features"]
        for feature in cur_features:
            for cfg in feature:
                if not cfg.endswith('_file'):
                    continue
                if feature[cfg] in self._custom_resources:
                    continue
                self._custom_resources.append(feature[cfg])
                if not odps.exist_resource(feature[cfg]):
                    print("[ERROR] resource file does not exist:", feature[cfg])
                    return False
                print("download resource file:", feature[cfg])
                with odps.open_resource(feature[cfg], mode='rb') as fp:
                    content=fp.read()
                    res_file = open(feature[cfg], "wb")
                    res_file.write(content)
                    res_file.close()

        pyfg.set_env('FEATURE_OPERATOR_DIR', os.getcwd())
        fg_handler = pyfg.FgHandler(features, 1)
        user_inputs = fg_handler.user_inputs()
        item_inputs = fg_handler.item_inputs()
        ctx_inputs = fg_handler.context_inputs()
        input_fields = user_inputs | item_inputs | ctx_inputs

        t = odps.get_table(self.input_table)
        input_schema = t.table_schema
        self._input_partitions = [pt.name for pt in input_schema.partitions]

        columns = []
        for column in input_schema.columns:
            if column.name in reserve_columns:
                columns.append(column)
                self._reserve_names.append(column.name)
                self._output_types.append(str(column.type))
                self._default_values.append(get_default_value(str(column.type)))
        for column in input_schema.columns:
            if column.name in input_fields:
                self._input_names.append(column.name)
                self._default_values.append(get_default_value(str(column.type)))

        fg_schema = fg_handler.table_schema()
        sorted_schema = sorted(fg_schema.items(), key=lambda x: x[0])
        for name, _type in sorted_schema:
            if name in reserve_columns:
                raise ValueError(f"feature name `{name}` in reserved columns, please rename it or delete the corresponding reserve column.")
            columns.append(Column(name=name, type=_type))
            self._output_types.append(_type)
            self._output_names.append(name)

        output_schema = TableSchema(columns=columns, partitions=input_schema.partitions)

        if odps.exist_table(self.output_table):
            output_t = odps.get_table(self.output_table)
            output_t_schema = output_t.table_schema
            if output_schema != output_t_schema:
                print("required output_schema != exist_table_schema")
                print("required output_schema = ", output_schema)
                print("exist_table_schema = ", output_t_schema)
                return False
        else:
            print(f"Create output table: {self.output_table}")
            options.sql.use_odps2_extension = True
            table = odps.create_table(self.output_table, output_schema, if_not_exists=True, lifecycle=t.lifecycle)
            return table is not None
        return True

    @property
    def reserve_names(self):
        return self._reserve_names

    @property
    def input_names(self):
        return self._input_names

    @property
    def default_values(self):
        return self._default_values

    @property
    def output_names(self):
        return self._output_names

    @property
    def output_types(self):
        return self._output_types

    @property
    def custom_resources(self):
        return self._custom_resources

    @property
    def input_partitions(self):
        return self._input_partitions