#!/usr/bin/env python
# -*- coding: utf-8 -*-
import os
import sys
import time
import logging
import hashlib
from string import Template
from odps import options
try:
  from .create_output_table import TableCreator
except:
  from create_output_table import TableCreator


class FgTask(object):
    def __init__(self, input_table, output_table, fg_json_file, partition_value=None, batch_size=128, force_delete_output_table=False, force_update_resource=False, debug=False):
        logging.basicConfig()
        logging.getLogger().setLevel(logging.INFO)
        self.input_table = input_table
        self.output_table = output_table
        self.partition_value = partition_value
        self.fg_json_file = fg_json_file
        self.force_delete_output_table = force_delete_output_table
        self.force_update_resource = force_update_resource
        self.batch_size = batch_size
        self.sql_settings = {}
        self.debug = debug
        self.work_path = os.path.abspath(os.path.dirname(__file__))
        self.version = self.get_version()
        self.fg_so_file = 'pyfg' + self.version + '.cpython-37m-x86_64-linux-gnu.so'
        self.init()

    def add_sql_setting(self, key, value):
        self.sql_settings[key] = value
        return self
   
    def get_version(self):
        version_path = os.path.join(self.work_path, 'VERSION')
        if os.path.exists(version_path):
            with open(version_path, 'r') as file:
                version = file.read().strip()
        else:
            version = ''
            logging.error("no fg version specified.")
        version_name = ''
        version_list = version.split('.')
        int_version_list = map(int, version_list)
        for i in int_version_list:
            if i > 9:
                raise ValueError("The version should not be more than 9.")
            version_name += str(i)
        logging.info("fg version: %s", version_name)
        os.environ['FG_VERSION'] = version_name
        return version_name

    def init(self):
        logging.info("input table: %s", self.input_table)
        logging.info("output table: %s", self.output_table)
        logging.info("partition value: %s", self.partition_value)
        logging.info("fg json file: %s", self.fg_json_file)
        logging.info("batch size: %d", self.batch_size)
        logging.info("force_delete_output_table: %s", self.force_delete_output_table)
        logging.info("force_update_resource: %s", self.force_update_resource)
        logging.info("sql settings: %s", self.sql_settings)
        logging.info("debug: %s", self.debug)
        if self.debug:
            self.force_update_resource = True
            logging.info("set force_update_resource=True in debug mode")
        self.output_table_creator = TableCreator(
            self.input_table,
            self.output_table, 
            self.fg_json_file, 
            self.force_delete_output_table)

    @staticmethod
    def add_odps_resource(odps, resource_name, resource_path, resource_type, force=False):
        if force and odps.exist_resource(resource_name):
            odps.delete_resource(resource_name)
            logging.info("delete resource: %s", resource_name)
        if not odps.exist_resource(resource_name):
            mode = 'r' if resource_type == 'py' else 'rb'
            fp = open(resource_path, mode)
            odps.create_resource(resource_name, resource_type, fileobj=fp)
            logging.info("create resource: %s", resource_name)

    @staticmethod
    def add_odps_function(odps, func, cls, res_names, update=False):
        res_names = set(res_names)
        logging.info("fg udtf resuorce list: %s", ','.join(res_names))
        resources = [odps.get_resource(res) for res in res_names]
        if not odps.exist_function(func):
            odps.create_function(func, class_type=cls, resources=resources)
        elif update:
            fn = odps.get_function(func)
            fn.class_type = cls
            fn.resources = resources
            fn.update()

    @staticmethod
    def gen_fg_udtf(args, filename):
        # read template file
        work_path = os.path.abspath(os.path.dirname(__file__))
        tpl_file = os.path.join(work_path, 'fg_udtf_template.py')
        with open(tpl_file, 'r', encoding='utf-8') as file:
            template_content = file.read()
        # create Template object
        template = Template(template_content)
        # do variable substitute
        result = template.safe_substitute(args)
        # save to a new file
        with open(filename, 'w', encoding='utf-8') as file:
            file.write(result)

    def run_odps_sql(self, odps, sql):
        options.sql.settings = { 
            "odps.sql.submit.mode": "script",
            "odps.sql.type.system.odps2": True,
            "odps.sql.python.version": "cp37"
        }
        instance = odps.run_sql(sql, hints=self.sql_settings)
        logging.info("instance ID = %s", instance.id)
        logging.info("logview: %s", instance.get_logview_address())
        while not instance.is_terminated():
            for task_name in instance.get_task_names():
                print(instance.get_task_progress(task_name).get_stage_progress_formatted_string())
            time.sleep(10)
        instance.wait_for_success()

    def __call__(self, odps, **kwargs):
        self.run(odps, **kwargs)

    def run(self, odps, **kwargs):
        if not self.output_table_creator(odps):
            logging.error("output table create failed")
            sys.exit(1)
        if self.debug:
            if self.output_table_creator.reserve_names:
                logging.info("output names: %s,%s", 
                    ','.join(self.output_table_creator.reserve_names),
                    ','.join(self.output_table_creator.output_names))
            else:
                logging.info("output names: %s", ','.join(self.output_table_creator.output_names))
            logging.info("output types: %s", ','.join(self.output_table_creator.output_types))

        fg_lib = os.path.join(self.work_path, 'pyfg.cpython-37m-x86_64-linux-gnu.so')
        FgTask.add_odps_resource(odps, self.fg_so_file, fg_lib, 'file', self.force_update_resource)
        loader_path = os.path.join(self.work_path, 'load_pyfg.py')
        FgTask.add_odps_resource(odps, 'load_pyfg.py', loader_path, 'py', self.force_update_resource)

        # 准备 udtf 需要的模板参数
        args = {'fg_json_file': self.fg_json_file, 'batch_size': self.batch_size}
        args['version'] = self.version
        args['output_type_list'] = self.output_table_creator.output_types
        args['output_types'] = ','.join(args['output_type_list'])
        args['reserve_names'] = self.output_table_creator.reserve_names
        args['input_names'] = self.output_table_creator.input_names
        args['default_values'] = self.output_table_creator.default_values
        args['output_names'] = self.output_table_creator.output_names
        args['debug'] = self.debug

        input_field_names = ','.join(args['input_names'])
        depends = [self.version, ','.join(args['output_names']), ','.join(args['output_types']),
            ','.join([str(x) for x in args['default_values']]), input_field_names,
            ','.join(self.output_table_creator.custom_resources)]
        depend_content = "__".join(depends).encode()
        md5 = hashlib.md5(depend_content).hexdigest()
        logging.info("task md5: %s", md5)

        udtf_py_file = f'fg_udtf_{md5}.py'
        FgTask.gen_fg_udtf(args, udtf_py_file)
        FgTask.add_odps_resource(odps, udtf_py_file, udtf_py_file, 'py', self.force_update_resource)

        udtf_func_name = f"fg_processor_{md5}"
        udtf_cls_name = f'fg_udtf_{md5}.FgProcess'
        resources = [self.fg_so_file, udtf_py_file, 'load_pyfg.py', self.fg_json_file]
        resources.extend(self.output_table_creator.custom_resources)
        FgTask.add_odps_function(odps, udtf_func_name, udtf_cls_name, resources)

        if args['reserve_names']:
            input_fields = ','.join(args['reserve_names']) + ',' + input_field_names
            output_fields = ','.join(args['reserve_names']) + ',' + ','.join(args['output_names'])
        else:
            input_fields = input_field_names
            output_fields = ','.join(args['output_names'])
        partitions = self.output_table_creator.input_partitions
        if not partitions:
            sql = f'''
            INSERT OVERWRITE TABLE {self.output_table} 
            SELECT {udtf_func_name}({input_fields})
            AS ({output_fields})
            from {self.input_table};'''
        else:
            if self.partition_value is None:
                raise ValueError("partition_value should not be None")
            pt_values = self.partition_value.split(',')
            where = []
            for pt, pt_value in zip(partitions, pt_values):
                where.append(pt + '=' + pt_value)
            select_partition = ' and '.join(where)
            insert_partition = ','.join(where)
            sql = f'''
            INSERT OVERWRITE TABLE {self.output_table} PARTITION({insert_partition})
            SELECT {udtf_func_name}({input_fields})
            AS ({output_fields})
            from {self.input_table} 
            where {select_partition};'''
        if self.debug:
            logging.info("sql: %s", sql)
        self.run_odps_sql(odps, sql)
        logging.info("Success")


if __name__ == '__main__':
    import os
    from odps import ODPS
    odps = ODPS(
        os.getenv('ALIBABA_CLOUD_ACCESS_KEY_ID'),
        os.getenv('ALIBABA_CLOUD_ACCESS_KEY_SECRET'),
        project='pai_rec_dev',
        endpoint='http://service.cn-beijing.maxcompute.aliyun.com/api',
    )
    fg_task = FgTask("tmp_text", "sentencepiece_output", "test_fg4.json")
    fg_task(odps)
